"""
    This file is part of the borb (R) project.
    Copyright (c) 2020-2040 borb Group NV
    Authors: Joris Schellekens, et al.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3
    as published by the Free Software Foundation with the addition of the
    following permission added to Section 15 as permitted in Section 7(a):
    FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY
    BORB GROUP. BORB GROUP DISCLAIMS THE WARRANTY OF NON INFRINGEMENT
    OF THIRD PARTY RIGHTS

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.

    See the GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program; if not, see http://www.gnu.org/licenses or write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA, 02110-1301 USA.

    The interactive user interfaces in modified source and object code versions
    of this program must display Appropriate Legal Notices, as required under
    Section 5 of the GNU Affero General Public License.
    In accordance with Section 7(b) of the GNU Affero General Public License,
    a covered work must retain the producer line in every PDF that is created
    or manipulated using borb.

    You can be released from the requirements of the license by purchasing
    a commercial license. Buying such a license is mandatory as soon as you
    develop commercial activities involving the borb software without
    disclosing the source code of your own applications.

    These activities include: offering paid services to customers as an ASP,
    serving PDFs on the fly in a web application, shipping borb with a closed
    source product.

    For more information, please contact borb Software Corp. at this
    address: joris.schellekens.1989@gmail.com
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Decodes data encoded in an ASCII base-85 representation,
reproducing the original binary data.
"""

import base64


class ASCII85Decode:
    """
    Decodes data encoded in an ASCII base-85 representation,
    reproducing the original binary data.
    """

    @staticmethod
    def decode(bytes_in: bytes) -> bytes:
        """
        Decodes data encoded in an ASCII base-85 representation
        """
        exceptions_to_throw = []

        # trivial case
        if len(bytes_in) == 0:
            return bytes_in

        # trimming
        if bytes_in[-1] == 10 and bytes_in[-2] == 13:
            bytes_in = bytes_in[0:-2]
        if bytes_in[-1] == 10:
            bytes_in = bytes_in[0:-1]
        if bytes_in[-1] == 13:
            bytes_in = bytes_in[0:-1]

        # normal decode
        try:
            return base64.a85decode(bytes_in)
        except Exception as e:
            exceptions_to_throw.append(e)
            pass

        # adobe decode
        try:
            return base64.a85decode(bytes_in, adobe=True)
        except Exception as e:
            exceptions_to_throw.append(e)
            pass

        # we should not be here
        raise exceptions_to_throw[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener extracts all text from a PDF Document
"""
import typing
from decimal import Decimal
from functools import cmp_to_key

from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.chunk_of_text_render_event import (
    ChunkOfTextRenderEvent,
    LeftToRightComparator,
)
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.page.page import Page


class SimpleTextExtraction(EventListener):
    """
    This implementation of EventListener extracts all text from a PDF Document
    """

    def __init__(self):
        self._text_render_info_per_page: typing.Dict[
            int, typing.List[ChunkOfTextRenderEvent]
        ] = {}
        self._text_per_page: typing.Dict[int, str] = {}
        self._current_page: int = -1

    def _event_occurred(self, event: Event) -> None:
        if isinstance(event, ChunkOfTextRenderEvent):
            self._render_text(event)
        if isinstance(event, BeginPageEvent):
            self._begin_page(event.get_page())
        if isinstance(event, EndPageEvent):
            self._end_page(event.get_page())

    def get_text_for_page(self, page_nr: int) -> str:
        """
        This function returns all text on a given page
        """
        return self._text_per_page[page_nr] if page_nr in self._text_per_page else ""

    def _render_text(self, text_render_info: ChunkOfTextRenderEvent):

        # init if needed
        if self._current_page not in self._text_render_info_per_page:
            self._text_render_info_per_page[self._current_page] = []

        # append TextRenderInfo
        self._text_render_info_per_page[self._current_page].append(text_render_info)

    def _begin_page(self, page: Page):
        self._current_page += 1

    def _end_page(self, page: Page):

        # get TextRenderInfo objects on page
        tris = (
            self._text_render_info_per_page[self._current_page]
            if self._current_page in self._text_render_info_per_page
            else []
        )

        # remove no-op
        tris = [x for x in tris if x._text is not None]
        tris = [x for x in tris if len(x._text.replace(" ", "")) != 0]

        # skip empty
        if len(tris) == 0:
            return

        # sort according to comparator
        tris = sorted(tris, key=cmp_to_key(LeftToRightComparator.cmp))

        # iterate over the TextRenderInfo objects to get the text
        last_baseline_bottom = tris[0].get_baseline().y
        last_baseline_right = tris[0].get_baseline().x
        text = ""
        for t in tris:

            # add newline if needed
            if abs(t.get_baseline().y - last_baseline_bottom) > 10 and len(text) > 0:
                if text.endswith(" "):
                    text = text[0:-1]
                text += "\n"
                text += t._text
                last_baseline_right = t.get_baseline().x + t.get_baseline().width
                last_baseline_bottom = t.get_baseline().y
                continue

            # check text
            if t._text.startswith(" ") or text.endswith(" "):
                text += t._text
                last_baseline_right = t.get_baseline().x + t.get_baseline().width
                continue

            # add space if needed
            delta = abs(last_baseline_right - t.get_baseline().x)
            space_width = round(t.get_space_character_width_estimate_in_user_space(), 1)
            text += " " if (space_width * Decimal(0.90) < delta) else ""

            # normal append
            text += t._text
            last_baseline_right = t.get_baseline().x + t.get_baseline().width
            continue

        # store text
        self._text_per_page[self._current_page] = text



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer acts as an aggregator for
its child transformers, allowing it to transform AnyPDFType
"""
import io
from typing import Optional, Union

from borb.io.read.types import AnyPDFType
from borb.io.write.ascii_art.ascii_art_transformer import ASCIIArtTransformer
from borb.io.write.document.catalog_transformer import CatalogTransformer
from borb.io.write.document.document_transformer import DocumentTransformer
from borb.io.write.document.information_dictionary_transformer import (
    InformationDictionaryTransformer,
)
from borb.io.write.image.image_transformer import ImageTransformer
from borb.io.write.object.array_transformer import ArrayTransformer
from borb.io.write.object.dictionary_transformer import DictionaryTransformer
from borb.io.write.object.stream_transformer import StreamTransformer
from borb.io.write.page.page_transformer import PageTransformer
from borb.io.write.page.pages_transformer import PagesTransformer
from borb.io.write.primitive.boolean_transformer import BooleanTransformer
from borb.io.write.primitive.name_transformer import NameTransformer
from borb.io.write.primitive.number_transformer import NumberTransformer
from borb.io.write.primitive.string_transformer import StringTransformer
from borb.io.write.reference.reference_transformer import ReferenceTransform
from borb.io.write.reference.xref_transformer import XREFTransformer
from borb.io.write.transformer import Transformer, WriteTransformerState
from borb.io.write.xmp.xmp_transformer import XMPTransformer


class AnyObjectTransformer(Transformer):
    """
    This implementation of WriteBaseTransformer acts as an aggregator for
    its child transformers, allowing it to transform AnyPDFType
    """

    def __init__(self):
        super().__init__()
        # fun
        self.add_child_transformer(ASCIIArtTransformer())
        # special object types
        self.add_child_transformer(DocumentTransformer())
        self.add_child_transformer(CatalogTransformer())
        self.add_child_transformer(XREFTransformer())
        self.add_child_transformer(PagesTransformer())
        self.add_child_transformer(PageTransformer())
        self.add_child_transformer(InformationDictionaryTransformer())
        # object types
        self.add_child_transformer(ArrayTransformer())
        self.add_child_transformer(StreamTransformer())
        self.add_child_transformer(DictionaryTransformer())
        self.add_child_transformer(ImageTransformer())
        self.add_child_transformer(XMPTransformer())
        # primitives
        self.add_child_transformer(NameTransformer())
        self.add_child_transformer(StringTransformer())
        self.add_child_transformer(ReferenceTransform())
        self.add_child_transformer(NumberTransformer())
        self.add_child_transformer(BooleanTransformer())

    def can_be_transformed(self, object_to_transform: AnyPDFType):
        """
        This function returns True if the object to be transformed
        can be transformed by this WriteBaseTransformer
        """
        return False

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
        destination: Optional[Union[io.BufferedIOBase, io.RawIOBase]] = None,
    ):
        """
        This method writes an (PDF) object to a byte stream
        """
        if context is None:
            super().transform(
                object_to_transform,
                WriteTransformerState(
                    destination=destination, root_object=object_to_transform
                ),
            )
        else:
            super().transform(object_to_transform, context)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Begin a text object, initializing the text matrix, Tm , and the text line matrix,
Tlm , to the identity matrix. Text objects shall not be nested; a second BT shall
not appear before an ET.
"""

import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.event.begin_text_event import BeginTextEvent
from borb.pdf.canvas.geometry.matrix import Matrix
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class BeginTextObject(CanvasOperator):
    """
    Begin a text object, initializing the text matrix, Tm , and the text line matrix,
    Tlm , to the identity matrix. Text objects shall not be nested; a second BT shall
    not appear before an ET.
    """

    def __init__(self):
        super().__init__("BT", 0)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the BT operator
        """
        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.text_matrix = Matrix.identity_matrix()
        canvas.graphics_state.text_line_matrix = Matrix.identity_matrix()
        for l in event_listeners:
            l._event_occurred(BeginTextEvent())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Paint the specified XObject. The operand name shall appear as a key in
the XObject subdictionary of the current resource dictionary (see 7.8.3,
"Resource Dictionaries"). The associated value shall be a stream whose
Type entry, if present, is XObject.
"""
import io
import typing

import PIL  # type: ignore [import]

from borb.io.read.types import AnyPDFType, Dictionary, Name, Stream
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class Do(CanvasOperator):
    """
    Paint the specified XObject. The operand name shall appear as a key in
    the XObject subdictionary of the current resource dictionary (see 7.8.3,
    "Resource Dictionaries"). The associated value shall be a stream whose
    Type entry, if present, is XObject.

    The effect of Do depends on the value
    of the XObject’s Subtype entry, which may be Image (see 8.9.5, "Image
    Dictionaries"), Form (see 8.10, "Form XObjects"), or PS (see 8.8.2,
    "PostScript XObjects").
    """

    def __init__(self):
        super().__init__("Do", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the Do operator
        """

        # get Page
        canvas = canvas_stream_processor.get_canvas()
        page = canvas_stream_processor.get_page()  # type: ignore [attr-defined]

        # get XObject
        assert isinstance(operands[0], Name)
        xobject = canvas_stream_processor.get_resource("XObject", str(operands[0]))

        # render Image objects
        if isinstance(xobject, PIL.Image.Image):
            for l in event_listeners:
                l._event_occurred(
                    ImageRenderEvent(
                        graphics_state=canvas.graphics_state, image=xobject
                    )
                )
            return

        # Form XObject
        if (
            isinstance(xobject, Stream)
            and "Subtype" in xobject
            and xobject["Subtype"] == "Form"
        ):

            # execute XObject
            xobject_resources: Dictionary = (
                xobject["Resources"] if "Resources" in xobject else {}
            )
            child_canvas_stream_processor = (
                canvas_stream_processor.create_child_canvas_stream_processor(
                    [xobject_resources]
                )
            )
            child_canvas_stream_processor.read(
                io.BytesIO(xobject["DecodedBytes"]), event_listeners
            )

            # return
            return

        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A Type 1 font program is a stylized PostScript program that describes glyph shapes. It uses a compact
encoding for the glyph descriptions, and it includes hint information that enables high-quality rendering even at
small sizes and low resolutions.
"""
import copy
import io
import logging
import typing
from pathlib import Path

from fontTools.afmLib import AFM  # type: ignore [import]
from fontTools.agl import toUnicode  # type: ignore [import]
from fontTools.cffLib import CFFFontSet, TopDict  # type: ignore [import]

from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, Name
from borb.pdf.canvas.font.adobe_standard_encoding import (
    adobe_standard_decode,
    adobe_standard_encode,
)
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.simple_font.simple_font import SimpleFont
from borb.pdf.canvas.font.symbol_encoding import symbol_decode, zapfdingbats_decode

logger = logging.getLogger(__name__)


class Type1Font(SimpleFont):
    """
    A Type 1 font program is a stylized PostScript program that describes glyph shapes. It uses a compact
    encoding for the glyph descriptions, and it includes hint information that enables high-quality rendering even at
    small sizes and low resolutions.
    """

    def __init__(self):
        super(Type1Font, self).__init__()
        self[Name("Type")] = Name("Font")
        self[Name("Subtype")] = Name("Type1")
        self._character_identifier_to_unicode_lookup: typing.Dict[int, str] = {}
        self._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {}

    def _read_encoding_with_unclear_glyph_names(self):

        # figure out how many characters we'll need to calculate
        # fmt: off
        assert "FirstChar" in self, "Type1Font must have a /FirstChar entry"
        assert isinstance(self["FirstChar"], bDecimal), "Type1Font must have a valid /FirstChar entry"
        assert "LastChar" in self, "Type1Font must have a /LastChar entry"
        assert isinstance(self["LastChar"], bDecimal), "Type1Font must have a valid /LastChar entry"
        # fmt: on
        first_char: int = int(self["FirstChar"])
        last_char: int = int(self["LastChar"])
        self._character_identifier_to_unicode_lookup = {}

        i: int = 0
        for i in range(first_char, last_char + 1):
            y: typing.Optional[str] = None
            try:
                if self["Encoding"]["BaseEncoding"] == "WinAnsiEncoding":
                    y = bytes([i]).decode("cp1252")
                elif self["Encoding"]["BaseEncoding"] == "MacRomanEncoding":
                    y = bytes([i]).decode("mac-roman")
                elif self["Encoding"]["BaseEncoding"] == "MacExpertEncoding":
                    # TODO replace by actual MacExpertEncoding
                    logger.debug(
                        "Font %s uses MacExpertEncoding, defaulting to MacRomanEncoding"
                        % str(self["BaseFont"])
                    )
                    y = bytes([i]).decode("mac-roman")
                elif self["Encoding"]["BaseEncoding"] == "StandardEncoding":
                    y = adobe_standard_decode(bytes([i]))
            except:
                pass
            if y is not None:
                self._character_identifier_to_unicode_lookup[i] = y

        # apply differences
        j: int = 0
        i = 0
        encoding_without_differences: typing.Dict[int, str] = copy.deepcopy(
            self._character_identifier_to_unicode_lookup
        )
        while i < len(self["Encoding"]["Differences"]):
            assert isinstance(self["Encoding"]["Differences"][i], bDecimal)
            character_code: int = self["Encoding"]["Differences"][i]
            j = i + 1
            while j < len(self["Encoding"]["Differences"]) and not isinstance(
                self["Encoding"]["Differences"][j], bDecimal
            ):
                glyph_name: str = str(self["Encoding"]["Differences"][j])
                cid: int = int(glyph_name[1:])
                self._character_identifier_to_unicode_lookup[
                    int(character_code)
                ] = encoding_without_differences[cid]
                character_code += 1
                j += 1
            i = j

        # build reverse map
        self._unicode_lookup_to_character_identifier = {
            v: k for k, v in self._character_identifier_to_unicode_lookup.items()
        }

    def _read_encoding_with_differences(self) -> None:

        # check whether we've been here before
        if len(self._unicode_lookup_to_character_identifier) > 0:
            return

        # IF self["Encoding"]["Differences"] only has non-standard glyph names,
        # AND those glyph names all resolve to <empty string> (using fontTools.agl.toUnicode)
        # THEN attempt to parse them as numbers, using those as CIDs
        if all(
            [
                str(x).startswith("G") and toUnicode(str(x)) == ""
                for x in self["Encoding"]["Differences"]
                if isinstance(x, Name)
            ]
        ):
            self._read_encoding_with_unclear_glyph_names()
            return

        # figure out how many characters we'll need to calculate
        # fmt: off
        assert "FirstChar" in self, "Type1Font must have a /FirstChar entry"
        assert isinstance(self["FirstChar"], bDecimal), "Type1Font must have a valid /FirstChar entry"
        assert "LastChar" in self, "Type1Font must have a /LastChar entry"
        assert isinstance(self["LastChar"], bDecimal), "Type1Font must have a valid /LastChar entry"
        # fmt: on
        first_char: int = int(self["FirstChar"])
        last_char: int = int(self["LastChar"])
        self._character_identifier_to_unicode_lookup = {}

        i: int = 0
        for i in range(first_char, last_char + 1):
            y: typing.Optional[str] = None
            try:
                if self["Encoding"]["BaseEncoding"] == "WinAnsiEncoding":
                    y = bytes([i]).decode("cp1252")
                elif self["Encoding"]["BaseEncoding"] == "MacRomanEncoding":
                    y = bytes([i]).decode("mac-roman")
                elif self["Encoding"]["BaseEncoding"] == "MacExpertEncoding":
                    # TODO replace by actual MacExpertEncoding
                    logger.debug(
                        "Font %s uses MacExpertEncoding, defaulting to MacRomanEncoding"
                        % str(self["BaseFont"])
                    )
                    y = bytes([i]).decode("mac-roman")
                elif self["Encoding"]["BaseEncoding"] == "StandardEncoding":
                    y = adobe_standard_decode(bytes([i]))
            except:
                pass
            if y is not None:
                self._character_identifier_to_unicode_lookup[i] = y

        # read font file
        if "FontDescriptor" in self and "FontFile3" in self["FontDescriptor"]:
            font_file_bytes: bytes = self["FontDescriptor"]["FontFile3"]["DecodedBytes"]
            cff: CFFFontSet = CFFFontSet()
            cff.major = 1
            cff.decompile(io.BytesIO(font_file_bytes), otFont=None)
            assert len(cff.keys()) == 1
            top_level_dict: TopDict = cff[0]
            # A Type 1 font program’s glyph descriptions are keyed by glyph names, not by character codes. Glyph names
            # are ordinary PDF name objects. Descriptions of Latin alphabetic characters are normally associated with
            # names consisting of single letters, such as A or a. Other characters are associated with names composed of
            # words, such as three, ampersand, or parenleft. A Type 1 font’s built-in encoding shall be defined by an
            # Encoding array that is part of the font program, not to be confused with the Encoding entry in the PDF font
            # dictionary.

        # apply differences
        j: int = 0
        i = 0
        while i < len(self["Encoding"]["Differences"]):
            assert isinstance(self["Encoding"]["Differences"][i], bDecimal)
            character_code: int = self["Encoding"]["Differences"][i]
            j = i + 1
            while j < len(self["Encoding"]["Differences"]) and not isinstance(
                self["Encoding"]["Differences"][j], bDecimal
            ):
                glyph_name: str = str(self["Encoding"]["Differences"][j])
                self._character_identifier_to_unicode_lookup[
                    int(character_code)
                ] = toUnicode(glyph_name)
                character_code += 1
                j += 1
            i = j

        # build reverse map
        self._unicode_lookup_to_character_identifier = {
            v: k for k, v in self._character_identifier_to_unicode_lookup.items()
        }

    def _read_to_unicode(self):
        if len(self._unicode_lookup_to_character_identifier) > 0:
            return
        # fmt: off
        assert "ToUnicode" in self, "Type1Font must have a /ToUnicode entry."
        assert ("DecodedBytes" in self["ToUnicode"]), "Type1Font must have a valid /ToUnicode entry."
        # fmt: on
        cmap_bytes: bytes = self["ToUnicode"]["DecodedBytes"]
        self._character_identifier_to_unicode_lookup = self._read_cmap(cmap_bytes)
        self._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {
            v: k for k, v in self._character_identifier_to_unicode_lookup.items()
        }

    def character_identifier_to_unicode(
        self, character_identifier: int
    ) -> typing.Optional[str]:
        """
        This function maps a character identifier to its unicode str.
        If no such mapping exists, this function returns None.
        """

        # If the font dictionary contains a ToUnicode CMap (see 9.10.3, "ToUnicode CMaps"), use that CMap to
        # convert the character code to Unicode.
        if Name("ToUnicode") in self:
            self._read_to_unicode()
            return self._character_identifier_to_unicode_lookup.get(
                character_identifier
            )

        # if "Encoding" is not present, the implied encoding is StandardEncoding
        if "Encoding" not in self:
            self[Name("Encoding")] = Name("StandardEncoding")
        if (
            "Encoding" in self
            and isinstance(self["Encoding"], Dictionary)
            and "BaseEncoding" not in self["Encoding"]
        ):
            self["Encoding"][Name("BaseEncoding")] = Name("WinAnsiEncoding")

        # If the font is a simple font that uses one of the predefined encodings MacRomanEncoding,
        # MacExpertEncoding, or WinAnsiEncoding,
        if isinstance(self["Encoding"], Name) and self["Encoding"] in [
            "MacRomanEncoding",
            "MacExpertEncoding",
            "WinAnsiEncoding",
            "StandardEncoding",
        ]:
            if character_identifier < 0 or character_identifier > 256:
                return None
            try:
                if self["Encoding"] == "WinAnsiEncoding":
                    return bytes([character_identifier]).decode("cp1252")
                elif self["Encoding"] == "MacRomanEncoding":
                    return bytes([character_identifier]).decode("mac-roman")
                elif self["Encoding"] == "MacExpertEncoding":
                    # TODO replace by actual MacExpertEncoding
                    logger.debug(
                        "Font %s uses MacExpertEncoding, defaulting to MacRomanEncoding"
                        % str(self["BaseFont"])
                    )
                    return bytes([character_identifier]).decode("mac-roman")
                elif self["Encoding"] == "StandardEncoding":
                    return adobe_standard_decode(bytes([character_identifier]))
                else:
                    logger.debug(
                        "Font %s uses unknown encoding %s"
                        % (str(self["BaseFont"]), str(self["Encoding"]))
                    )
            except UnicodeDecodeError:
                return None

        # or that has an encoding whose Differences array includes
        # only character names taken from the Adobe standard Latin character set and the set of named characters
        # in the Symbol font (see Annex D)
        # a) Map the character code to a character name according to Table D.1 and the font’s Differences
        # array.
        # b) Look up the character name in the Adobe Glyph List (see the Bibliography) to obtain the
        # corresponding Unicode value.
        if (
            isinstance(self["Encoding"], Dictionary)
            and "BaseEncoding" in self["Encoding"]
            and self["Encoding"]["BaseEncoding"]
            in [
                "MacRomanEncoding",
                "MacExpertEncoding",
                "WinAnsiEncoding",
                "StandardEncoding",
            ]
        ):
            self._read_encoding_with_differences()
            if character_identifier < 0 or character_identifier > 256:
                return None
            return self._character_identifier_to_unicode_lookup.get(
                character_identifier
            )

        # default
        return None

    def unicode_to_character_identifier(self, unicode: str) -> typing.Optional[int]:
        """
        This function maps a unicode str to its character identifier.
        If no such mapping exists, this function returns None.
        """
        if Name("ToUnicode") in self:
            self._read_to_unicode()
            return self._unicode_lookup_to_character_identifier.get(unicode)

        # if "Encoding" is not present, the implied encoding is StandardEncoding
        if "Encoding" not in self:
            self[Name("Encoding")] = Name("StandardEncoding")

        if isinstance(self["Encoding"], Name) and self["Encoding"] in [
            "MacRomanEncoding",
            "MacExpertEncoding",
            "WinAnsiEncoding",
            "StandardEncoding",
        ]:
            try:
                if self["Encoding"] == "WinAnsiEncoding":
                    return int(unicode.encode("cp1252"))
                elif self["Encoding"] == "MacRomanEncoding":
                    return int(unicode.encode("mac-roman"))
                elif self["Encoding"] == "MacExpertEncoding":
                    # TODO replace by actual MacExpertEncoding
                    return int(unicode.encode("mac-roman"))
                elif self["Encoding"] == "StandardEncoding":
                    return int(adobe_standard_encode(unicode))
            except:
                return None

        if (
            isinstance(self["Encoding"], Dictionary)
            and "BaseEncoding" in self["Encoding"]
            and self["Encoding"]["BaseEncoding"]
            in [
                "MacRomanEncoding",
                "MacExpertEncoding",
                "WinAnsiEncoding",
                "StandardEncoding",
            ]
        ):
            self._read_encoding_with_differences()
            return self._unicode_lookup_to_character_identifier.get(unicode, None)

        # default
        return None

    def get_width(self, character_identifier: int) -> typing.Optional[bDecimal]:
        """
        This function returns the width (in text space) of a given character identifier.
        If this Font is unable to represent the glyph that corresponds to the character identifier,
        this function returns None
        """
        first_char: int = int(self["FirstChar"])
        last_char: int = int(self["LastChar"])
        if first_char <= character_identifier <= last_char:
            return self["Widths"][character_identifier - first_char]
        return None

    def get_ascent(self) -> bDecimal:
        """
        This function returns the maximum height above the baseline reached by glyphs in this font.
        The height of glyphs for accented characters shall be excluded.
        """
        return self["FontDescriptor"]["Ascent"]

    def get_descent(self) -> bDecimal:
        """
        This function returns the maximum depth below the baseline reached by glyphs in this font.
        The value shall be a negative number.
        """
        return self["FontDescriptor"]["Descent"]

    def _empty_copy(self) -> "Font":
        return Type1Font()

    def __deepcopy__(self, memodict={}):
        # fmt: off
        f_out: Font = super(Type1Font, self).__deepcopy__(memodict)
        f_out[Name("Subtype")] = Name("Type1")
        f_out._character_identifier_to_unicode_lookup: typing.Dict[int, str] = {k: v for k, v in self._character_identifier_to_unicode_lookup.items()}
        f_out._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {k: v for k, v in self._unicode_lookup_to_character_identifier.items()}
        return f_out
        # fmt: on


class StandardType1Font(Type1Font):
    """
    The PostScript names of 14 Type 1 fonts, known as the standard 14 fonts, are as follows: Times-Roman,
    Helvetica, Courier, Symbol, Times-Bold, Helvetica-Bold, Courier-Bold, ZapfDingbats, Times-Italic, Helvetica-
    Oblique, Courier-Oblique, Times-BoldItalic, Helvetica-BoldOblique, Courier-BoldOblique
    These fonts, or their font metrics and suitable substitution fonts, shall be available to the conforming reader.
    """

    STANDARD_14_FONT_NAMES: typing.List[str] = [
        "Courier",
        "Courier-Bold",
        "Courier-Bold-Oblique",
        "Courier-Oblique",
        "Helvetica",
        "Helvetica-Bold",
        "Helvetica-Bold-Oblique",
        "Helvetica-Oblique",
        "Symbol",
        "Times-Bold",
        "Times-Bold-Italic",
        "Times-Italic",
        "Times-Roman",
        "ZapfDingbats",
    ]

    @staticmethod
    def _canonical_name(font_name: str) -> typing.Optional[str]:
        def _to_lower_and_alpha(x: str) -> str:
            return "".join([c for c in x.lower() if c in "abcdefghijklmnopqrstuvwxyz"])

        canonical_name: str = _to_lower_and_alpha(font_name)
        for n in StandardType1Font.STANDARD_14_FONT_NAMES:
            if _to_lower_and_alpha(n) == canonical_name:
                return n

        return None

    @staticmethod
    def is_standard_14_font_name(font_name: str) -> bool:
        """
        This function returns True if the given str represents the name of one of the standard 14 fonts, False otherwise
        """
        return StandardType1Font._canonical_name(font_name) is not None

    # fmt: off
    def __init__(self, font_name: typing.Optional[str] = None):
        super(StandardType1Font, self).__init__()
        if font_name is not None:

            font_name = StandardType1Font._canonical_name(font_name)
            assert font_name is not None, "font_name must be one of the 14 StandardType1Font names."

            # check whether AFM directory exists
            afm_directory: Path = Path(__file__).parent / "afm"
            assert afm_directory.exists(), "AFM directory not found"

            # check whether AFM file exists
            afm_file: Path = afm_directory / (font_name.lower() + ".afm")
            assert afm_file.exists(), "afm file not found"

            # build AFM datastructure
            self._afm: AFM = AFM(afm_file)

            self[Name("Type")] = Name("Font")
            self[Name("Subtype")] = Name("Type1")
            self[Name("BaseFont")] = Name(self._afm._attrs["FontName"])

            self._character_identifier_to_unicode_lookup: typing.Dict[int, str] = {}
            self._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {}

            if font_name == "Symbol":
                self._character_identifier_to_unicode_lookup  = {c:symbol_decode(bytes([c])) for c in range(0, 256)}
                self._unicode_lookup_to_character_identifier = {v:k for k,v in self._character_identifier_to_unicode_lookup.items()}

            elif font_name == "ZapfDingbats":
                self._character_identifier_to_unicode_lookup = {c:zapfdingbats_decode(bytes([c])) for c in range(0, 256)}
                self._unicode_lookup_to_character_identifier = {v:k for k,v in self._character_identifier_to_unicode_lookup.items()}

            else:
                self[Name("Encoding")] = Name("WinAnsiEncoding")
                for c in range(0, 256):
                    try:
                        self._character_identifier_to_unicode_lookup[c] = bytes([c]).decode("cp1252")
                    except:
                        self._character_identifier_to_unicode_lookup[c] = ""
                self._unicode_lookup_to_character_identifier = {v:k for k,v in self._character_identifier_to_unicode_lookup.items()}

    # fmt: on

    def character_identifier_to_unicode(
        self, character_identifier: int
    ) -> typing.Optional[str]:
        """
        This function maps a character identifier to its unicode str.
        If no such mapping exists, this function returns None.
        """
        return self._character_identifier_to_unicode_lookup.get(character_identifier)

    def unicode_to_character_identifier(self, unicode: str) -> typing.Optional[int]:
        """
        This function maps a unicode str to its character identifier.
        If no such mapping exists, this function returns None.
        """
        return self._unicode_lookup_to_character_identifier.get(unicode)

    def get_width(self, character_identifier: int) -> typing.Optional[bDecimal]:
        """
        This function returns the width (in text space) of a given character identifier.
        If this Font is unable to represent the glyph that corresponds to the character identifier,
        this function returns None
        """
        widths: typing.List[bDecimal] = [
            bDecimal(v[1])
            for k, v in self._afm._chars.items()
            if v[0] == character_identifier
        ]
        if len(widths) == 1:
            return widths[0]
        return bDecimal(0)

    def get_ascent(self) -> bDecimal:
        """
        This function returns the maximum height above the baseline reached by glyphs in this font.
        The height of glyphs for accented characters shall be excluded.
        """
        if "Ascender" in self._afm._attrs:
            return bDecimal(self._afm._attrs["Ascender"])
        return bDecimal(0)

    def get_descent(self) -> bDecimal:
        """
        This function returns the maximum depth below the baseline reached by glyphs in this font.
        The value shall be a negative number.
        """
        if "Descender" in self._afm._attrs:
            return bDecimal(self._afm._attrs["Descender"])
        return bDecimal(0)

    def _empty_copy(self) -> "Font":
        return StandardType1Font()

    def __deepcopy__(self, memodict={}):
        # fmt: off
        f_out: Font = super(StandardType1Font, self).__deepcopy__(memodict)
        f_out[Name("Subtype")] = Name("Type1")
        f_out._character_identifier_to_unicode_lookup: typing.Dict[int, str] = {k: v for k, v in self._character_identifier_to_unicode_lookup.items()}
        f_out._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {k: v for k, v in self._unicode_lookup_to_character_identifier.items()}
        f_out._afm = self._afm
        return f_out
        # fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of ReadBaseTransformer is responsible for reading CCITT fax images
"""
import io
import logging
import typing
from typing import Any, Optional, Union

from PIL import Image  # type: ignore [import]

from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, Stream, add_base_methods
from borb.pdf.canvas.event.event_listener import EventListener

logger = logging.getLogger(__name__)


class CCITTFaxImageTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading CCITT fax images
    """

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be transformed is a CCITT Image
        """
        return (
            isinstance(object, Stream)
            and object.get("Type", None) in ["XObject", None]
            and object.get("Subtype", None) == ("Image")
            and "Filter" in object
            and (
                object["Filter"] == "CCITTFaxDecode"
                or (
                    isinstance(object["Filter"], list)
                    and object["Filter"][0] == "CCITTFaxDecode"
                )
            )
        )

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a CCITT Image from a byte stream
        """
        # use PIL to read image bytes
        # fmt: off
        assert isinstance(object_to_transform, Stream), "object_to_transform must be of type Stream"
        # fmt: on
        try:
            tmp = Image.open(io.BytesIO(object_to_transform["Bytes"]))
            tmp.getpixel(
                (0, 0)
            )  # attempting to read pixel 0,0 will trigger an error if the underlying image does not exist
        except:
            logger.debug(
                "Unable to read ccitt fax image. Constructing empty image of same dimensions."
            )
            w = int(object_to_transform["Width"])
            h = int(object_to_transform["Height"])
            tmp = Image.new("RGB", (w, h), (128, 128, 128))

        # add base methods
        add_base_methods(tmp)

        # set parent
        tmp.set_parent(parent_object)  # type: ignore[attr-defined]

        # return
        return tmp



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Farrow & Ball is a British manufacturer of paints and wallpapers largely based upon historic colour palettes and archives.
The company is particularly well known for the unusual names of its products.
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HexColor, RGBColor


class FarrowAndBall(HexColor):
    """
    Farrow & Ball is a British manufacturer of paints and wallpapers largely based upon historic colour palettes and archives.
    The company is particularly well known for the unusual names of its products.
    """

    COLOR_DEFINITION = {
        "all-white": "#faf9ef",
        "pointing": "#f9f3e3",
        "james-white": "#ece9d7",
        "white-tie": "#f2ead6",
        "clunch": "#e5dcc8",
        "blackened": "#dfdfdb",
        "tallow": "#f5ead2",
        "dimity": "#ece2d2",
        "great-white": "#e6e1dc",
        "strong-white": "#e9e7dd",
        "slipper-satin": "#eae4d3",
        "shaded-white": "#dad4c2",
        "new-white": "#f4e9d0",
        "lime-white": "#e9dfc9",
        "ringworld-ground": "#efe0c5",
        "matchstick": "#e6d8be",
        "house-white": "#f2e7c9",
        "off-white": "#e0d6c0",
        "string": "#dcceb0",
        "savage-ground": "#d7c5a9",
        "cord": "#d6c39f",
        "cream": "#e3cea7",
        "hay": "#d8c194",
        "light-stone": "#d1c1a0",
        "joas-white": "#e2d4be",
        "archive": "#decbb3",
        "stony-ground": "#cdc2ad",
        "cornforth-white": "#d4cfc7",
        "elephants-breath": "#cdc3b7",
        "lamp-room-gray": "#b7b6ae",
        "light-blue": "#b9beb5",
        "hardwick-white": "#b6b1a1",
        "blue-gray": "#b5b7a7",
        "french-gray": "#b4b19a",
        "pigeon": "#9fa195",
        "card-room-green": "#899081",
        "old-white": "#cdc3ad",
        "bone": "#cdc4ae",
        "fawn": "#ccbfa5",
        "biscuit": "#bca783",
        "light-gray": "#b2a894",
        "mouses-back": "#968975",
        "smoked-trout": "#bea895",
        "dead-salmon": "#b09b8c",
        "london-stone": "#b2a18d",
        "buff": "#b59e82",
        "drab": "#9b8a6f",
        "dauphin": "#928069",
        "pink-ground": "#ecd8ca",
        "calamine": "#e8d6d1",
        "setting-plaster": "#dbc2b0",
        "dutch-pink": "#d5b08a",
        "entrance-hall-pink": "#c39e80",
        "flower-pink": "#d6a787",
        "pintment-pink": "#c7997e",
        "red-earth": "#bb7b69",
        "porphyry-pink": "#b17b6c",
        "fox-red": "#a8725d",
        "loggia": "#aa6d5b",
        "blazer": "#ac514c",
        "book-room-red": "#a3695b",
        "picture-gallery-red": "#975b4f",
        "rectory-red": "#99434c",
        "radiocchio": "#8f4a51",
        "etruscan-red": "#7e534a",
        "eating-room-red": "#864e4f",
        "pale-hound": "#e6ddb8",
        "hound-lemon": "#e2d2a4",
        "farrows-cream": "#e8d4b0",
        "dayroom-yellow": "#f2dc9e",
        "dorset-cream": "#eacfa2",
        "gervase-yellow": "#dcc98f",
        "yellow-ground": "#efca89",
        "citron": "#eecc81",
        "ciara-yellow": "#e1c17e",
        "straw": "#d7b37e",
        "print-room-yellow": "#dbb678",
        "babouche": "#eabe67",
        "sudbury-yellow": "#d9b475",
        "octagon-yellow": "#d1ae72",
        "orangery": "#e0ae6c",
        "india-yellow": "#c09759",
        "cane": "#caa26c",
        "sand": "#ba9166",
        "green-ground": "#dbd8b6",
        "cooking-apple-green": "#c4c3a2",
        "vert-de-terre": "#babca4",
        "ball-green": "#bab393",
        "stone-white": "#b3ac8c",
        "green-stone": "#a59c78",
        "lichen": "#9e9f87",
        "olive": "#8b896b",
        "sutcliffe-green": "#899677",
        "pea-green": "#849779",
        "calke-green": "#77866a",
        "minster-green": "#586552",
        "powder-blue": "#acbdb4",
        "chappell-green": "#93a595",
        "saxon-green": "#99a680",
        "folly-green": "#95aa86",
        "breakfast-room-green": "#94a588",
        "arsenic": "#88b49b",
        "pale-powder": "#dde1d5",
        "borrowed-light": "#d8dfdc",
        "skylight": "#c6ccc7",
        "teresas-green": "#c2cfc2",
        "green-blue": "#aebdb1",
        "castle-gray": "#8d988c",
        "blue-ground": "#a7c9c8",
        "dix-blue": "#9ab1ab",
        "ballroom-blue": "#8faba7",
        "sugar-bag-light": "#81a09a",
        "oval-room-blue": "#8b9d9b",
        "berrington-blue": "#7e9495",
        "parma-gray": "#b1c0c2",
        "lulworth-blue": "#a2b8c7",
        "stone-blue": "#77989f",
        "chinese-blue": "#6f96a6",
        "cooks-blue": "#6f94b3",
        "pitch-blue": "#61718e",
        "wainscot": "#7e634e",
        "down-pipe": "#626665",
        "hague-blue": "#3e4e56",
        "brinjal": "#5a4348",
        "mahogany": "#524947",
        "off-black": "#424548",
        "green-smoke": "#717b70",
        "monkey-puzzle": "#4b5853",
        "caggiage-green": "#434c47",
        "railings": "#43474a",
        "studio-green": "#474d4a",
        "black-blue": "#3d4345",
    }

    def __init__(self, color_name: str):
        assert color_name in FarrowAndBall.COLOR_DEFINITION
        self.color_name: str = color_name
        super(FarrowAndBall, self).__init__(FarrowAndBall.COLOR_DEFINITION[color_name])

    def get_name(self) -> str:
        """
        This function returns the name of this `Farrow and Ball` color
        """
        return self.color_name

    @staticmethod
    def find_nearest_farrow_and_ball_color(color: Color) -> "FarrowAndBall":
        """
        This function find the nearest `Farrow and Ball` equivalent for a given Color
        """
        rgb_color_001: RGBColor = color.to_rgb()
        d_min: typing.Optional[Decimal] = None
        c_min: typing.Optional[str] = None
        for n, c in FarrowAndBall.COLOR_DEFINITION.items():
            rgb_color_002: RGBColor = HexColor(c)
            d: Decimal = (
                (rgb_color_001.red - rgb_color_002.red) ** 2
                + (rgb_color_001.green - rgb_color_002.green) ** 2
                + (rgb_color_001.blue - rgb_color_002.blue) ** 2
            )
            if d_min is None or d < d_min:
                d_min = d
                c_min = n
        assert d_min is not None
        assert c_min is not None
        return FarrowAndBall(c_min)

    def __deepcopy__(self, memodict={}):
        return FarrowAndBall(self.color_name)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener enables you to
    extract font-related information from a Document
"""
import typing
from typing import List

from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.font.font import Font


class FontExtraction(EventListener):
    """
    This implementation of EventListener enables you to
    extract font-related information from a Document
    """

    def __init__(self):
        self._fonts_per_page: typing.Dict[int, typing.List[Font]] = {}
        self._current_page: int = -1

    def _event_occurred(self, event: Event) -> None:
        if isinstance(event, BeginPageEvent):
            self._begin_page(event)

    def _begin_page(self, event: BeginPageEvent):

        # update page number
        self._current_page += 1
        self._fonts_per_page[self._current_page] = []

        # get page
        page = event.get_page()
        if page is None:
            return

        # get resources
        if "Resources" not in page or not isinstance(page["Resources"], dict):
            return
        if "Font" not in page["Resources"] or not isinstance(
            page["Resources"]["Font"], dict
        ):
            return

        for _, f in page["Resources"]["Font"].items():
            self._fonts_per_page[self._current_page].append(f)

    def get_fonts_for_page(self, page_number: int) -> List[Font]:
        """
        This function returns all fonts used on a given page
        """
        return (
            self._fonts_per_page[page_number]
            if page_number in self._fonts_per_page
            else []
        )

    def get_font_names_for_page(self, page_number: int) -> List[str]:
        """
        This function returns all font names used on a given page
        """
        out: typing.List[str] = []
        for x in self.get_fonts_for_page(page_number):
            font_name: typing.Optional[str] = x.get_font_name()
            if font_name is not None:
                out.append(font_name)
        return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    A conforming reader shall maintain an internal data structure called the graphics state that holds current
    graphics control parameters. These parameters define the global framework within which the graphics
    operators execute.
"""
import copy
import typing
from decimal import Decimal

from borb.io.read.types import Name
from borb.pdf.canvas.color.color import RGBColor
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.geometry.matrix import Matrix


class CanvasGraphicsState:
    """
    A conforming reader shall maintain an internal data structure called the graphics state that holds current
    graphics control parameters. These parameters define the global framework within which the graphics
    operators execute.

    EXAMPLE 1
    The f (fill) operator implicitly uses the current colour parameter, and the S (stroke) operator additionally
    uses the current line width parameter from the graphics state.

    A conforming reader shall initialize the graphic state at the beginning of each page with the values specified in
    Table 52 and Table 53. Table 52 lists those graphics state parameters that are device-independent and are
    appropriate to specify in page descriptions. The parameters listed in Table 53 control details of the rendering
    (scan conversion) process and are device-dependent; a page description that is intended to be device-
    independent should not be written to modify these parameters.
    """

    def __init__(self):
        self.ctm: Matrix = Matrix.identity_matrix()
        self.text_matrix: Matrix = Matrix.identity_matrix()
        self.text_line_matrix: Matrix = Matrix.identity_matrix()
        self.text_rise: Decimal = Decimal(0)
        self.character_spacing: Decimal = Decimal(0)
        self.word_spacing: Decimal = Decimal(0)
        self.horizontal_scaling: Decimal = Decimal(100)
        self.leading: Decimal = Decimal(0)
        self.font: typing.Optional[typing.Union["Name", "Font"]] = None
        self.font_size: Decimal = Decimal(0)
        self.path: typing.List["LineSegment"] = []
        self.clipping_path: typing.List["LineSegment"] = []
        self.non_stroke_color_space = None
        self.non_stroke_color = RGBColor(Decimal(0), Decimal(0), Decimal(0))
        self.stroke_color_space = None
        self.stroke_color = RGBColor(Decimal(0), Decimal(0), Decimal(0))
        self.line_width = Decimal(1)
        self.line_cap = None
        self.line_join = None
        self.miter_limit = Decimal(10)
        self.dash_pattern = None
        self.rendering_intent = None
        self.stroke_adjustment = None
        self.blend_mode = None
        self.soft_mask = None
        self.alpha_constant = None
        self.alpha_source = None

    def __deepcopy__(self, memodict={}):
        out = CanvasGraphicsState()

        # ctm
        out.ctm = copy.deepcopy(self.ctm, memodict)

        # text matrix
        if self.text_matrix is not None:
            out.text_matrix = copy.deepcopy(self.text_matrix, memodict)

        # text line matrix
        if self.text_line_matrix is not None:
            out.text_line_matrix = copy.deepcopy(self.text_line_matrix, memodict)

        out.text_rise = self.text_rise
        out.character_spacing = self.character_spacing
        out.word_spacing = self.word_spacing
        out.horizontal_scaling = self.horizontal_scaling
        out.leading = self.leading
        if self.font is not None:
            if isinstance(self.font, Font):
                out.font = self.font.__deepcopy__(memodict)
            if isinstance(self.font, Name):
                out.font = copy.deepcopy(self.font)
        out.font_size = self.font_size
        # out.clipping_path = None
        out.non_stroke_color_space = copy.deepcopy(self.non_stroke_color_space)
        out.non_stroke_color = copy.deepcopy(self.non_stroke_color)
        out.stroke_color_space = copy.deepcopy(self.stroke_color_space)
        out.stroke_color = copy.deepcopy(self.stroke_color)
        out.line_width = self.line_width
        # self.line_cap = None
        # self.line_join = None
        out.miter_limit = self.miter_limit
        # self.dash_pattern = None
        # self.rendering_intent = None
        # self.stroke_adjustment = None
        # self.blend_mode = None
        # self.soft_mask = None
        # self.alpha_constant = None
        # self.alpha_source = None
        return out



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This module contains the basis for events and eventlisteners
"""


class Event:
    """
    This class represents a generic event
    """

    pass


class EventListener:
    """
    This class represents a generic event listener
    This listener is notified whenever the canvas processes an event/command.
    """

    def _event_occurred(self, event: Event) -> None:
        """
        This method is called whenever a matching Event is fired from the Canvas.
        EventListeners can then choose to act on those Event objects.
        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class attempts to optimize a Type0Font by performing subsetting.
Subsetting is performed using fontTools.subset, after which the dictionaries are updated in the
Type0Font, the page content is also changed, updating the TJ instructions to refer to the new GIDs.
"""
import io
import random
import typing
import zlib

from fontTools.subset import Subsetter as fSubsetter  # type: ignore[import]
from fontTools.ttLib import TTFont  # type: ignore[import]

from borb.io.read.types import Decimal as bDecimal, Name
from borb.io.write.font.character_set_listener import CharacterSetListener
from borb.io.write.font.copy_command_operator import CopyCommandOperator
from borb.io.write.font.subset_show_text_with_glyph_positioning import (
    SubSetShowTextWithGlyphPositioning,
)
from borb.pdf import Page
from borb.pdf.canvas.canvas import Canvas
from borb.pdf.canvas.canvas_stream_processor import CanvasStreamProcessor
from borb.pdf.canvas.font.composite_font.font_type_0 import Type0Font
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.simple_font.true_type_font import TrueTypeFont


class Subsetter:
    """
    This class attempts to optimize a Type0Font by performing subsetting.
    Subsetting is performed using fontTools.subset, after which the dictionaries are updated in the
    Type0Font, the page content is also changed, updating the TJ instructions to refer to the new GIDs.
    """

    @staticmethod
    def _extract_text_per_font(page: "Page") -> typing.Dict[Font, typing.Set[str]]:
        csp = CanvasStreamProcessor(page, Canvas(), [])
        l: CharacterSetListener = CharacterSetListener()
        csp.read(io.BytesIO(page["Contents"]["DecodedBytes"]), [l])
        return l.get_character_set_per_font()

    @staticmethod
    def _modify_page_content_stream(
        page: Page, old_fonts: typing.List[Font], new_fonts: typing.List[Font]
    ):

        # build CanvasStreamProcessor
        csp = CanvasStreamProcessor(page, Canvas(), [])

        # new content stream
        new_content_stream: bytearray = bytearray()

        # modify operators
        for k, v in csp._canvas_operators.items():
            csp._canvas_operators[k] = CopyCommandOperator(v, new_content_stream)

        # rewrite the operands for Tf
        csp._canvas_operators["TJ"] = SubSetShowTextWithGlyphPositioning(
            old_fonts, new_fonts, new_content_stream
        )

        # process page
        csp.read(io.BytesIO(page["Contents"]["DecodedBytes"]))

        # modify Contents
        page["Contents"][Name("DecodedBytes")] = new_content_stream
        page["Contents"][Name("Bytes")] = zlib.compress(new_content_stream, 9)
        page["Contents"][Name("Filter")] = Name("FlateDecode")
        page["Contents"][Name("Length")] = bDecimal(len(new_content_stream))

        # return
        return page

    @staticmethod
    def apply(page: Page) -> Page:
        """
        This function applies Font subsetting to the input Page and returns the Page afterwards
        :param page:    the Page on which to apply Font subsetting
        :return:        the Page with Font subsetting
        """
        if "Resources" not in page:
            return page
        if not isinstance(page["Resources"], dict):
            return page
        if "Font" not in page["Resources"]:
            return page

        # determine which fonts to apply subsetting to
        fonts_to_be_subset: typing.List[Type0Font] = [
            x for x in page["Resources"]["Font"].values() if isinstance(x, Type0Font)
        ]
        if len(fonts_to_be_subset) == 0:
            return page

        # determine character set for each Type0Font
        characters_per_font: typing.Dict[
            Font, typing.Set[str]
        ] = Subsetter._extract_text_per_font(page)

        # update which fonts to apply subsetting to
        fonts_to_be_subset = [
            x for x in fonts_to_be_subset if len(characters_per_font[x]) < 256
        ]
        if len(fonts_to_be_subset) == 0:
            return page

        subset_fonts: typing.List[Font] = []
        for old_font in fonts_to_be_subset:

            # get DecodedBytes
            font_file_bytes_001 = old_font["DescendantFonts"][0]["FontDescriptor"][
                "FontFile2"
            ]["DecodedBytes"]
            ttfont: TTFont = TTFont(io.BytesIO(font_file_bytes_001))

            # perform subsetting
            subsetter = fSubsetter()
            subsetter.populate(text="".join([x for x in characters_per_font[old_font]]))
            subsetter.options.glyph_names = True
            subsetter.options.recalc_bounds = True
            subsetter.options.recalc_average_width = True
            subsetter.subset(ttfont)

            # determine (modified) DecodedBytes
            with io.BytesIO() as font_file_bytes_002:
                ttfont.save(font_file_bytes_002)
                font_file_bytes_002 = font_file_bytes_002.getvalue()

            # build TrueTypeFont
            # fmt: off
            new_font: TrueTypeFont = TrueTypeFont.true_type_font_from_file(font_file_bytes_002)
            basefont_prefix: str = "".join([random.choice("ABCDEFGHIJKLMNOPQRSTUVWXYZ") for _ in range(0, 6)])
            new_font[Name("BaseFont")] = Name(basefont_prefix + "+" + str(new_font[Name("BaseFont")]))
            new_font["FontDescriptor"][Name("FontName")] = new_font[Name("BaseFont")]
            new_font[Name("ToUnicode")] = TrueTypeFont._build_custom_cmap_for_type_0_font(ttfont)
            new_font.pop(Name("Encoding"))

            # fmt: on
            subset_fonts.append(new_font)

        # change Page / Contents
        Subsetter._modify_page_content_stream(page, fonts_to_be_subset, subset_fonts)

        # change Page / Resources / Font
        for k, old_font in page["Resources"]["Font"].items():
            if old_font not in fonts_to_be_subset:
                continue
            new_font: Font = subset_fonts[fonts_to_be_subset.index(old_font)]
            page["Resources"]["Font"][k] = new_font

        # return
        return page
